CON              
                                                           
  _CLKMODE = XTAL1 + PLL16X                              
  _XINFREQ = 5_000_000

  CLK_FREQ = ((_clkmode - xtal1) >> 6) * _xinfreq
  MS_001   = CLK_FREQ / 1_000
  US_001   = CLK_FREQ / 1_000_000
  PSEUDO_MULTIPLIER = 1000 '' For pseudo real math, also referred to as scaled integer math.
  BRIGHTNESS = 255 / 3  ' Use 255 for maximum brightness.
  WSDI = 7     'prop port for di pin on ws2812's
        
  LEDS_IN_RING = 16
  RINGS_IN_USE = 3
  TOTAL_LEDS = LEDS_IN_RING * RINGS_IN_USE
  HALF_RING = LEDS_IN_RING / 2
  MAX_HALF_INDEX = HALF_RING - 1
  MAX_RING_INDEX = LEDS_IN_RING - 1
  MAX_LED_INDEX = TOTAL_LEDS - 1
  MAX_RING_ROTATION_AMOUNT = LEDS_IN_RING / 2
  MAX_ROTATION_AMOUNT = LEDS_IN_RING / 2
  MAX_DIFFERENCE = 255 / 2
  MIN_DIFFERENCE = 255 / 6
  MAX_SPEED = 100
  MIN_SPEED = 80 
       
VAR
  
  long stripRotationTimer[RINGS_IN_USE], stripRotationInterval[RINGS_IN_USE], stripSpeed[RINGS_IN_USE]
  long stripDelay[RINGS_IN_USE], differenceDirection[RINGS_IN_USE], speedDirection[RINGS_IN_USE]
  long speedChangeTimer[RINGS_IN_USE], speedChangeInterval[RINGS_IN_USE]
  long ringAddress[RINGS_IN_USE], positionChange[RINGS_IN_USE]
  
OBJ

  strip : "jm_ws2812"
                   
PUB Main | addr, localIndex, difference[RINGS_IN_USE], initialColorIndex[RINGS_IN_USE], {
  } finalColorIndex[RINGS_IN_USE], initialColor, finalColor, position[RINGS_IN_USE]
    
  strip.start(WSDI, TOTAL_LEDS)
    
  addr := strip.address
 
  repeat localIndex from 0 to RINGS_IN_USE - 1
    ringAddress[localIndex] := addr + ((4 * LEDS_IN_RING) * localIndex)
    initialColorIndex[localIndex] := localIndex * (255 / RINGS_IN_USE)
    differenceDirection[localIndex] := 1
    speedDirection[localIndex] := 1
    difference[localIndex] := MIN_DIFFERENCE + (localIndex * ((MAX_DIFFERENCE - {
      } MIN_DIFFERENCE) / RINGS_IN_USE))
    stripSpeed[localIndex] := MIN_SPEED + (localIndex * ((MAX_SPEED - {
      } MIN_SPEED) / RINGS_IN_USE))
    stripDelay[localIndex] := speedToDelay[stripSpeed[localIndex]]
    stripRotationInterval[localIndex] := MS_001 * stripDelay[localIndex]
    speedChangeInterval[localIndex] := clkfreq / 10 
    position[localIndex] := localIndex
    positionChange[localIndex] := 1

  positionChange[1] := -1
  speedDirection[2] := -1

  repeat 0 '2   ' fill rings with rainbow
    FillWheels
    waitcnt(clkfreq * 2 + cnt)
    FillWheelsDim(BRIGHTNESS)
    waitcnt(clkfreq * 2 + cnt)
  
  repeat localIndex from 0 to RINGS_IN_USE - 1 
    speedChangeTimer[localIndex] := cnt 
    stripRotationTimer[localIndex] := cnt
        
  repeat
 
    repeat localIndex from 0 to RINGS_IN_USE - 1

      if CheckRingTime(localIndex)
        finalColorIndex[localIndex] := initialColorIndex[localIndex] + difference[localIndex] 
        if initialColorIndex[localIndex] > 255 
          initialColorIndex[localIndex] -= 255
          positionChange[localIndex] *= -1
        if finalColorIndex[localIndex] > 255 
          finalColorIndex[localIndex] -= 255
        difference[localIndex] += differenceDirection[localIndex]
        if difference[localIndex] => MAX_DIFFERENCE
          differenceDirection[localIndex] := -1
        elseif difference[localIndex] =< MIN_DIFFERENCE
          differenceDirection[localIndex] := 1
        initialColor := WheelDim(initialColorIndex[localIndex], BRIGHTNESS)
        finalColor := WheelDim(finalColorIndex[localIndex], BRIGHTNESS)  
        Mirror(initialColor, finalColor, ringAddress[localIndex], position[localIndex])
        initialColorIndex[localIndex]++
        position[localIndex] += positionChange[localIndex]
        if  position[localIndex] > MAX_RING_INDEX
          position[localIndex] -= LEDS_IN_RING
        elseif position[localIndex] < 0
          position[localIndex] += LEDS_IN_RING
          
PUB FillWheels

  repeat result from 0 to RINGS_IN_USE - 1
    FillWheel(result)

PUB FillWheelsDim(localBrightness)

  repeat result from 0 to RINGS_IN_USE - 1
    FillWheelDim(result, localBrightness)

PUB FillWheel(wheelIndex) | firstLed

  firstLed := wheelIndex * LEDS_IN_RING
  
  repeat result from 0 to MAX_RING_INDEX 
    strip.set(result + firstLed, Wheel(result * (255 / LEDS_IN_RING)))
    waitcnt(clkfreq / 10 + cnt)

PUB FillWheelDim(wheelIndex, localBrightness) | firstLed

  firstLed := wheelIndex * LEDS_IN_RING
  
  repeat result from 0 to MAX_RING_INDEX 
    strip.set(result + firstLed, WheelDim(result * (255 / LEDS_IN_RING), localBrightness))
    waitcnt(clkfreq / 10 + cnt)

PUB Wheel(pos)

'' Creates color from 0 to 255 position input
'' -- colors transition r-g-b back to r

  if (pos < 85)
    return strip.color(255-pos*3, pos*3, 0)
  elseif (pos < 170)
    pos -= 85
    return strip.color(0, 255-pos*3, pos*3)
  else
    pos -= 170
    return strip.color(pos*3, 0, 255-pos*3)

PUB WheelDim(pos, localBrightness)

'' Creates color from 0 to 255 position input
'' -- colors transition r-g-b back to r
'' This version also adjusts the brightness of the LED.
'' Color resolution is lost when using the brightness
'' feature.

  if (pos < 85)
    return strip.colorx(255-pos*3, pos*3, 0, localBrightness)
  elseif (pos < 170)
    pos -= 85
    return strip.colorx(0, 255-pos*3, pos*3, localBrightness)
  else
    pos -= 170
    return strip.colorx(pos*3, 0, 255-pos*3, localBrightness)

PUB MergeColors(initialColor, finalColor, distinationPtr, localSize) | tempColor[3], {
    } colorMultiplier, localIndex, initialColors[3], finalColors[3], pseudoIncrements[3]
'' localSize should include the initialColor but not finalColor.
  
  BustColors(initialColor, @initialColors)
  BustColors(finalColor, @finalColors)
   
  ' find incremental color change
  ' use the "PSEUDO_MULTIPLIER" mulitplier to reduce rounding errors
  ' Note we divide by multiplier prior to assigning color
  ' to expanded pattern buffer.
  repeat result from 0 to 2   
    pseudoIncrements[result] := ((finalColors[result] - initialColors[result])* PSEUDO_MULTIPLIER) / localSize 
  
  colorMultiplier := 1
   
  ' assign initial color
  long[distinationPtr] := initialColor
   
  ' assign between colors
  ' final color not copied here. It will be copied as the
  ' initial color of the next loop
  repeat localSize - 1
    distinationPtr += 4
    repeat result from 0 to 2
      tempColor[result] := initialColors[result] + ((pseudoIncrements[result] * {
                         } colorMultiplier) / PSEUDO_MULTIPLIER)
    ' assign altered color
    long[distinationPtr] := Strip.Color(tempColor[0], tempColor[1], tempColor[2])
    colorMultiplier++


PUB BustColors(localColor, localPtr)
'' Split the a color value of "localColor" into its components parts and store these values
'' in three sequential longs located at "localPtr".

  long[localPtr] := localColor.byte[2]
  long[localPtr + 4] := localColor.byte[1]
  long[localPtr + 8] := localColor.byte[0]
 
PUB Rotate(localPtr, localSize, localRotation) | nearEdgeAddress, farEdgeAddress, {
                                               } tempBuffer[MAX_ROTATION_AMOUNT]
'' Shifts the buffer of "localSize" longs at memory location "localPtr" by "localRotation"
'' number of longs.
'' Not used in this program.

  localRotation   := -MAX_ROTATION_AMOUNT #> localRotation <# MAX_ROTATION_AMOUNT
  
  nearEdgeAddress := localPtr + (4 * ||localRotation)
  farEdgeAddress := localPtr + (4 * (localSize - ||localRotation))
  if localRotation > 0
    longmove(@tempBuffer, farEdgeAddress, localRotation)
    longmove(nearEdgeAddress, localPtr, localSize - localRotation)
    longmove(localPtr, @tempBuffer, localRotation)    '
  else
    ||localRotation 
    longmove(@tempBuffer, localPtr, localRotation)
    longmove(localPtr, nearEdgeAddress, localSize - localRotation)
    longmove(farEdgeAddress, @tempBuffer, localRotation)

PUB CheckRingTime(localIndex)
'' The "stripSpeed" variable will be changed by value of "speedDirection"
'' if an interval of time "speedChangeInterval" has transpired from
'' the last speed change.

  CheckStripSpeed(localIndex) ' check to see if the stripRotationInterval should be changed

  if cnt - stripRotationTimer[localIndex] < stripRotationInterval[localIndex]
    return 0
  else  
    stripRotationTimer[localIndex] += stripRotationInterval[localIndex]
    result := 1
    
PUB CheckStripSpeed(localIndex)
'' The "stripSpeed" variable will be changed by value of "speedDirection"
'' if an interval of time "speedChangeInterval" has transpired from
'' the last speed change.

  if cnt - speedChangeTimer[localIndex] < speedChangeInterval[localIndex]
    return
  else        
    stripSpeed[localIndex] += speedDirection[localIndex]
    if stripSpeed[localIndex] => MAX_SPEED
      speedDirection[localIndex] := -1
    elseif stripSpeed[localIndex] =< MIN_SPEED
      speedDirection[localIndex] := 1     
    stripDelay[localIndex] := speedToDelay[stripSpeed[localIndex]]
    stripRotationInterval[localIndex] := MS_001 * stripDelay[localIndex] 
    speedChangeTimer[localIndex] += speedChangeInterval[localIndex]
    
PUB Mirror(initialColor, finalColor, localPtr, position) | localIndex, scratchBuffer[HALF_RING]
'' Fill long buffer at "localPtr" with a mirrored colors.

  repeat while position > MAX_RING_INDEX
    position -= LEDS_IN_RING
  MergeColors(initialColor, finalColor, @scratchBuffer, HALF_RING)
 
  result := position - 1
  repeat localIndex from 0 to MAX_HALF_INDEX
    if result < 0
      result += LEDS_IN_RING
    if position > MAX_RING_INDEX
      position -= LEDS_IN_RING
             
    long[localPtr][position++] := scratchBuffer[localIndex]
    long[localPtr][result--] := scratchBuffer[localIndex]  
 
DAT 'speed to delay lookup table, values are milliseconds
'' Lookup table from another project. This project would benefit from a different table.

speedToDelay            long 2000, 1951, 1904, 1857, 1811, 1765, 1721, 1677, 1635, 1593
                        long 1551, 1511, 1471, 1432, 1394, 1357, 1320, 1284, 1249, 1214
                        long 1180, 1147, 1115, 1083, 1051, 1021, 991, 961, 933, 904
                        long 877, 850, 824, 798, 772, 748, 723, 700, 677, 654
                        long 632, 610, 589, 569, 549, 529, 510, 491, 473, 455
                        long 438, 421, 404, 388, 373, 358, 343, 328, 314, 301
                        long 287, 275, 262, 250, 238, 227, 216, 205, 194, 184
                        long 175, 165, 156, 147, 139, 131, 123, 115, 108, 101
                        long 94, 88, 82, 76, 70, 65, 59, 55, 50, 46
                        long 41, 37, 34, 30, 27, 24, 21, 18, 16, 14
extraJustIncase         long 12, 10, 8, 6, 4, 3, 2, 1, 0
   